package ca.recrown.islandsurvivalcraft.world.generation;

import java.util.LinkedList;
import java.util.Random;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;

import org.bukkit.Chunk;
import org.bukkit.Location;
import org.bukkit.Material;
import org.bukkit.World;
import org.bukkit.block.Biome;
import org.bukkit.event.EventHandler;
import org.bukkit.event.EventPriority;
import org.bukkit.event.Listener;
import org.bukkit.event.world.ChunkLoadEvent;
import org.bukkit.generator.ChunkGenerator;

import ca.recrown.islandsurvivalcraft.Utilities;
import ca.recrown.islandsurvivalcraft.caching.Cache;
import ca.recrown.islandsurvivalcraft.datatypes.Point2;
import ca.recrown.islandsurvivalcraft.world.BiomeSelector;
import ca.recrown.islandsurvivalcraft.world.IslandWorldMapper;
import ca.recrown.islandsurvivalcraft.world.shaders.WorldHeightShader;
import ca.recrown.islandsurvivalcraft.world.shaders.WorldLayerShader;

public class IslandWorldChunkGenerator extends ChunkGenerator implements Listener {
    private final Cache<Point2, Double> blockValueCache = new Cache<>(131072);
    private final Cache<Point2, Biome[]> biomeCache = new Cache<>(131072);
    private final Cache<Point2, Boolean> chunkExistenceCache = new Cache<>(131072);
    private final BiomeSelector biomeSelector = new BiomeSelector();
    private final ExecutorService exAlpha = Utilities.ISC_EXECUTOR_ALPHA;
    private final ExecutorService exBeta = Utilities.ISC_EXECUTOR_BETA;
    private volatile World currentWorld;

    public IslandWorldChunkGenerator() {
        biomeSelector.initialize();
    }

    @Override
    public ChunkData generateChunkData(World world, Random random, int chunkX, int chunkZ, BiomeGrid biomeGrid) {
        this.currentWorld = world;
        long seed = world.getSeed();
        int maxHeight = world.getMaxHeight();
        int seaLevel = world.getSeaLevel();
        IslandWorldMapper mapper = new IslandWorldMapper(seed, blockValueCache);
        TemperatureMapGenerator temperatureMapGenerator = new TemperatureMapGenerator(seed);
        WorldHeightShader heightShader = new WorldHeightShader(seed, mapper, seaLevel, world.getMaxHeight(), 3);
        WorldLayerShader layerShader = new WorldLayerShader(seed, seaLevel, maxHeight, mapper);
        BiomeGenerator biomeGenerator = new UniBiomeIslandGenerator();

        LinkedList<Future<Boolean>> tasks = new LinkedList<>();
        ChunkData chunkData = createChunkData(world);
        Future<Boolean> preLoader = exAlpha.submit(() -> {
            for (int x = Utilities.CHUNK_SIZE - 1; x >= 0; x--) {
                for (int z = Utilities.CHUNK_SIZE - 1; z >= 0; z--) {
                    if (Thread.currentThread().isInterrupted()) return false;
                    mapper.getWorldValue(Utilities.CHUNK_SIZE * chunkX + x, Utilities.CHUNK_SIZE * chunkZ + z);
                }
            }
            return true;
        });

        Biome[][] biomes = new Biome[Utilities.CHUNK_SIZE][Utilities.CHUNK_SIZE];
        for (int x = 0; x < Utilities.CHUNK_SIZE; x++) {
            for (int z = 0; z < Utilities.CHUNK_SIZE; z++) {
                final int localX = x;
                final int localZ = z;
                if (biomes[localX][localZ] == null) {
                    biomeGenerator.generateBiomeColumn(biomes, world, chunkX, chunkZ, localX, localZ, mapper, biomeSelector,
                            temperatureMapGenerator, biomeCache, chunkExistenceCache, random);
                }
                if (biomes[localX][localZ] == null) throw new IllegalStateException("Biome was null.");
                tasks.add(exBeta.submit(() -> {
                    for (int y = 0; y < maxHeight; y++) {
                        biomeGrid.setBiome(localX, y, localZ, biomes[localX][localZ]);
                    }
                    return true;
                }));

                final int worldX = Utilities.CHUNK_SIZE * chunkX + localX;
                final int worldZ = Utilities.CHUNK_SIZE * chunkZ + localZ;
                Biome currentBiome = biomes[localX][localZ];
                int terrainHeight = heightShader.getTerrainHeight(worldX, worldZ, currentBiome);
                int currentTerrainHeight = terrainHeight;
                int bedrockHeight = random.nextInt(5) + 1;
                if (layerShader.hasSpecialLayers(currentBiome)) {
                    Material currentMaterial = layerShader.getMaterialForHeight(worldX, worldZ, currentTerrainHeight, currentBiome);
                    while (currentMaterial != null) {
                        chunkData.setBlock(localX, currentTerrainHeight, localZ, currentMaterial);
                        currentTerrainHeight --;
                    }
                } else {
                    int surfaceThickness = layerShader.getSurfaceThickness(worldX, worldZ, currentBiome);
                    currentTerrainHeight -= surfaceThickness;
                    chunkData.setRegion(localX, currentTerrainHeight, localZ, localX + 1, currentTerrainHeight + surfaceThickness, localZ + 1, layerShader.getSurfaceMaterial(currentBiome));
                    int transitionThickness = layerShader.getTransitionMaterialThickness(worldX, worldZ, currentBiome);
                    currentTerrainHeight -= transitionThickness;
                    chunkData.setRegion(localX, currentTerrainHeight, localZ, localX + 1, currentTerrainHeight + transitionThickness, localZ + 1, layerShader.getTransitionMaterial(currentBiome));
                }
                chunkData.setRegion(localX, bedrockHeight, localZ, localX + 1, currentTerrainHeight + 1, localZ + 1, Material.STONE);

                if (terrainHeight < seaLevel) {
                    chunkData.setRegion(localX, terrainHeight, localZ, localX + 1, seaLevel, localZ + 1, Material.WATER);
                }
                chunkData.setRegion(localX, 0, localZ, localX + 1, bedrockHeight, localZ + 1, Material.BEDROCK);
            }
        }

        preLoader.cancel(false);
        try {
            while (!tasks.isEmpty()) {
                tasks.poll().get();
            }
        } catch (ExecutionException | InterruptedException e) {
            e.printStackTrace();
            throw new IllegalStateException(e.getCause().getMessage());
        }
        return chunkData;
    }

    @Override
    public Location getFixedSpawnLocation(World world, Random random) {
        Location location = new Location(world, 0, 128, 0);
        return location;
    }


    @Override
    public boolean isParallelCapable() {
        return true;
    }

    @EventHandler(priority = EventPriority.MONITOR)
    public void onChunkLoaded(ChunkLoadEvent event) {
        if (event.isNewChunk()) {
            if (currentWorld == event.getWorld()) {
                Chunk chunk = event.getChunk();
                
                chunkExistenceCache.set(new Point2(chunk.getX(), chunk.getZ()), true);
            }
        }
    }
}